#!/usr/bin/perl
##
## Meta.pm
##
## $Id: Meta.pm,v 1.51.2.3 2000/08/23 17:19:37 jarmbrecht Exp $ 
## Copyright (c) 1999-2000 Cobalt Networks, Inc.
##
## +++++++
## SUMMARY
## +++++++
##
## simple object interface to a postgreSQL database
##   (all "objects" have a type, refered to by the "type" trait)
## transaction rollback support for failed object save attempts
## import/export objects to XML format
##
## ++++++++++++++++
## GLOBAL FUNCTIONS
## ++++++++++++++++
##
## &connect($dbname, $user, $port);
##
##    connects to the postgres database $dbname as $user on $port
##
## &set_encoding($encoding);
##
##    sets the multibyte client encoding method, set based upon the value
##    in /etc/cobalt/locale when the database is first connected
##
## &get_encoding();
##
##    returns the current encoding method, displaying it on STDERR
##
## &db_reset();
##
##    verifies the database handle and resets if necessary
##
## ++++++++++++
## TRANSACTIONS
## ++++++++++++
##
## &begin_txn($ID);
##
##    initiates a user transaction in chained mode,
##    all user statements will be executed in a single transaction
##    until an explicit COMMIT, ROLLBACK, or execution abort is made
##    $ID is unique identifier for this transaction -- for nested
##
## &commit_txn($ID);
##
##    commits the current transaction,
##    all changes made by the transaction become visible to
##    others and are guaranteed to be durable if a crash occurs
##    $ID is unique identifier for this transaction -- for nested
##
## &rollback_txn($ID);
##
##    rolls back the current transaction,
##    causes all the updates made by the transaction to be discarded
##    $ID is unique identifier for this transaction -- for nested
##
## +++++++
## OBJECTS
## +++++++
##
## all "objects" have a type, refered to by the "type" trait.
## they must be first be created with their "type":
##
##    $obj = Cobalt::Meta->new( %arg );
##    $obj = Cobalt::Meta->new( "type"   => $type );
##    $obj = Cobalt::Meta->new( "type"   => $type,
##                              "name"   => $name,
##                              "modify" => $modify, ... );
## 
## they can be populated with values from the database using the retrieve
## function and providing a value for the "name" key and any additional conditions
##
##    $obj->retrieve("name", [ <$conj, $left, $op, $right>, ... ]);
##
## to get the value of a key use the get function:
##
##    $value = $obj->get($key);
##
## to obtain a locale-formatted date/time string of the last
## modification date of the meta object use
##
##    $modify = $obj->getmodify();
##
## returns the object type:
##
##    $type = $obj->gettype();
##
## returns the object name:
##
##    $type = $obj->getname();
##
## to insert a key=>value pair to the object:
##
##    $obj->put($key => $value);
##
## to create a new table in the database:
## with the attributes of object $obj
##
##    $obj->create_table();
##
## to save your object changes to the database:
##
##    $obj->save();
##
## to delete objects (only local copy):
##
##    $obj->DESTROY();
##
## to delete objects from the database:
##
##    $obj->delete();
##
## you can do more complex SQL queries:
## (this will return an array of objects)
##
##    &query("type"     => <meta type>,
##           "table"    => <into new table>,
##           "sort"     => <sort field>,
##           "order"    => "asc" | "dsc",
##           "distinct" => 1 | 0,
##           "range"    => [ <low, high> ],
##           "where"    => [ <left, op, right>,
##                           <conj, left, op, right>, ... ]);
##
## you can make updates directly to the database:
##
##    &update("type"   => table,
##            "change" => [ <key, value>, ... ],
##            "where"  => [ <left, op, right>,
##                          <conj, left, op, right>, ... ]);
##
## you can add fields to tables:
##
##    &add_field("type"  => $table,
##               "fname" => $key,
##               "ftype" => $type);
##
## for debugging purposes there is a display function that will simply
## output the key=>value pairs to STDERR
##
## output is formatted with &meta_write() calls.
##
##    $obj->display();
##
## +++++++++++++
## XML FUNCTIONS
## +++++++++++++
##
## you may dump objects as XML.
##
##    $xml = $obj->as_xml();
##
## and create them from XML.
##
##    $obj = Cobalt::Meta->from_xml_elem( $xml );
##
## ++++++++++++++++
## META INHERITANCE
## ++++++++++++++++
##
## objects may have members, these only exist while the object is created
## and must be eplicitly populated for retrieved objects.
##
## add childen objects with
##
##    $obj->add_member( $child_object );
##
## retrieve child objects with
##
##    @children = $obj->populate_members();
##
## To populate the children of a retrieved object use
##
##    $obj->populate_members();
##
## to see the members of an object (Not put them into the object.)
## use get_members. Args are the same as for query, but will only
## get you objects that are members and of the type specified.
##
##    @members = $obj->get_members( ... );
##

##------------------------------------------------------------------------
package Cobalt::Meta;
##

use vars qw(@ISA @EXPORT);
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw($METABUG);

use Pg qw(PGRES_EMPTY_QUERY
	  PGRES_COMMAND_OK
	  PGRES_CONNECTION_OK
	  PGRES_TUPLES_OK
	  PGRES_COPY_OUT
	  PGRES_COPY_IN
	  PGRES_BAD_RESPONSE
	  PGRES_NONFATAL_ERROR
	  PGRES_FATAL_ERROR);
use Pg;
use Cobalt::Util;
use Cobalt::Locale;
use Time::Local;
use XML::Parser;
use MIME::Base64;
use Fcntl;
use POSIX qw(locale_h strftime tmpnam);
use vars qw($dbc $Meta_txn $Meta_global_txn $Meta_id %client_encodings %pg_month);

$METABUG = 0;
$Meta_txn = '';
$Meta_global_txn = "global";

#
## ---------------------------------------------------------------------
#

# Meta configuration variables
my $Meta_dbname = "cobalt";
my $Meta_dbuser = "admin";
my $Meta_dbport = 5583;
$Meta_id = "/etc/cobalt/.meta.id";

# list of encoding types, referenced by their LANGUAGE
%client_encodings = ("en_US"      => "LATIN1",  # English/US
		     "ja_JP.sjis" => "SJIS",    # Japanese SJIS
		     "ja_JP.SJIS" => "SJIS",    # Japanese SJIS
		     "ja_JP.ujis" => "EUC_JP",  # Japanese EUC
		     "ja_JP.EUC"  => "EUC_JP",  # Japanese EUC
		     "ja_JP"      => "EUC_JP",  # Japanese EUC
		     "ja"         => "EUC_JP",  # Japanese EUC
		     "zh_CN"      => "EUC_CN",  # Chinese EUC
		     "ko_KP"      => "EUC_KR",  # Korean EUC
		     "unicode"    => "UNICODE",
		     "ascii"      => "SQL_ASCII");

%pg_month = (Jan => 0, Feb => 1, Mar => 2, Apr => 3,
	     May => 4, Jun => 5, Jul => 6, Aug => 7,
	     Sep => 8, Oct => 9, Nov => 10, Dec => 11);

#
## ---------------------------------------------------------------------
#

# set the database handle

use locale;

my $localeObj = Cobalt::Locale->new();
$dbc = &connect($Meta_dbname, $Meta_dbuser, $Meta_dbport);

if (defined($dbc)) {
    &set_encoding($client_encodings{ $localeObj->get_locale() });
}

no locale;

#
## ---------------------------------------------------------------------
#

#### Object deiniftion section.

# Maybe this should be put into a sperate file. But parsing that would
# be a little too much overhead and I'm worried about making the module 
# slower already.

# You can define commiters, children and files for an object.
# Commuters take the object and make it real.
# Children define which other objects are the children of it.
# Files define which files belong to an object when it's dumped
# to a cobalt-object file. 

# Ins and outs of how cobalt-object files handle the uids and gids
# are yet to be defined.

# These subroutines are used for commiting objects to "reality"
# the support a new type of object add a refrence to a function
# which takes an object of this type as the sole argument and
# makes your dreams come true.
#
# Note that the responsibility for adding children of the object lies
# with these functions so that for example if we were adding the site
# we would be able to the group name for the users depending on
# any mods we had to make to the site.
#
# In the future these functions should also handle modification as well as
# addition.
my %write_object_methods = ("vsite" => \&Cobalt::Vsite::vsite_obj_add,
			    "users" => \&Cobalt::User::site_user_add);

# If a type can have children you should define a series of hashes
# for it. The first element is the type of child, the second is the qualify
# statment to find the children.
#
# Any strings of the form [self.\w+] will be replaced by the value of that field
# in the parent object.
my %member_find = ("vsite" => { "users" => ["vsite", "=", "[self.name]"],
				"bw"	=> ["name", "=", "[self.ipaddr]"] });

# Objects may have data associated with it that is not held in the database.
# for example the web content of a virtual site, the files and home directory
# of a user.
#
# Again, this is a hash or refrences to hashes where any [self.\w+]
# occurences will be turned into their value from the database.
#
# The key of a hash is it's identifier. So that the 'webdata' files of a
# virtual site may be in a completly differnt place in a future product or even
# in the situation where RaQ users may be moved to a qube and visa versa.

my $vroot = "/home/sites";
my $mroot = "/var/spool/mail";
my %files = ( "vsite" => { "webdata" => "$vroot/[self.name]/web/",
                           "logs"    => "$vroot/[self.name]/logs/",
                           "certs"   => "$vroot/[self.name]/certs/" },
              "users" => { "data"    => "$vroot/[self.vsite]/users/[self.name]/",
                           "mail"    => "$mroot/[self.name]" });

############### new()
#
# syntax:       $obj = Cobalt::Meta->new("type" => $type);
#               returns an empty object belonging to database $type
#
# syntax:       $obj = Cobalt::Meta->new("type" => $type,
#                                        "key"  => $value, ...);
#               returns an object with corresponding ("key" => $value) traits
#
sub new
{
    my ($self,%args) = (@_);

    unless (defined($self)) {
	warn "META new - ERROR: object self undefined\n";
	return;
    }

    unless (exists($args{type})) {
	warn "META new - ERROR: attribute [self->type] undefined\n";
	return;
    }

    my $obj = bless {
	"type" => $args{type}
    }, $self;

    while(my($key,$value) = each %args)
    {
	$obj->{$key} = $value;
    }

    return $obj;
}

############### DESTROY()
#
# syntax:       $obj->DESTROY();
#               object destructor
#
sub DESTROY
{
    my $self = shift;
    return;
}

############### connect()
#
# syntax:       Cobalt::Meta::connect("dbname", "username", "dbport");
#               connects to the postgres database using passed arguments
#
#               database name = $dbname
#               database user = $pguser
#               database port = $pgport
#
sub connect
{
    my ($dbname, $pguser, $pgport) = @_;
    my $pass = (-f $Meta_id) ? qx(cat $Meta_id) : "";

    my $command = "user=$pguser port=$pgport dbname=$dbname password=$pass";

    warn "META connect - QUERY: $command\n" if ($METABUG);
    $dbc = Pg::connectdb($command);

    unless ($dbc->status eq PGRES_CONNECTION_OK)
    {
	warn "META connect - FAIL: " . $dbc->errorMessage . "\n" if ($METABUG);
	return;
    }

    if ($METABUG)
    {
	my $tfile;
	do {
	    $tfile = tmpnam();
	} until sysopen(TRACEFH, $tfile, O_WRONLY|O_CREAT|O_EXCL, 0600);

	$dbc->trace(*TRACEFH);

	warn "postgreSQL TraceFile: $tfile\n";
	warn "postgreSQL UserName:  $pguser\n";
	warn "postgreSQL Database:  $dbname\n";
	warn "postgreSQL Port:      $pgport\n";
	warn "META connect - OK\n";
    }

    return $dbc;
}

############### set_encoding
#
# syntax:       &set_encoding($client_encoding);
#               sets the CLIENT_ENCODING variable 
#
sub set_encoding
{
    my $encoding = shift || "SQL_ASCII";
    my $command = "SET CLIENT_ENCODING TO '$encoding'";

    warn "META set_encoding - QUERY: $command\n" if ($METABUG);
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
	warn "META set_encoding - FAIL: " . $dbc->errorMessage . "\n";
	return;
    }

    return 1;
}

############### get_encoding
#
# syntax:       $encoding = &get_encoding();
#               returns the CLIENT_ENCODING variable 
#
sub get_encoding
{
    &db_reset($dbc);

    my $command = "SHOW CLIENT_ENCODING;";
    warn "META get_encoding - QUERY: $command\n" if ($METABUG);
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
	warn "META get_encoding - FAIL: " . $dbc->errorMessage . "\n";
	return 0;
    }

    return 1;
}

############### db_reset()
#
# syntax:       &db_reset($dbc);
#               reset the database handle if necessary
#
sub db_reset
{
    my $connection = shift;

    return unless (defined($connection));
    $connection->reset unless ($connection->status eq PGRES_CONNECTION_OK);
    return $connection;
}

############### begin_txn("transaction ID")
#
# syntax:       &begin_txn($id)
#               begin a user transaction
#
sub begin_txn
{
    my($txn_id) = (shift || $Meta_global_txn);

    unless ($Meta_txn eq '')
    {
	warn "META BEGIN transaction ($txn_id) - ERROR: transaction already in progress\n" if ($METABUG);
	return FALSE;
    }

    &db_reset($dbc);

    my $result = $dbc->exec("BEGIN TRANSACTION");
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
	warn "META BEGIN transaction ($txn_id) - FAIL: " . $dbc->errorMessage . "\n";
	return FALSE;
    }

    warn "META BEGIN transaction ($txn_id) - OK\n" if ($METABUG);
    $Meta_txn = $txn_id;

    return TRUE;
}

############### commit_txn("transaction ID")
#
# syntax:       &commit_txn($id)
#               commit a transaction
#
sub commit_txn
{
    my($txn_id) = (shift || $Meta_global_txn);

    if ($Meta_txn eq '')
    {
	warn "META COMMIT transaction ($txn_id) - ERROR: no transaction in progress\n" if ($METABUG);
	return FALSE;
    }
    unless ($Meta_txn eq $txn_id)
    {
	warn "META COMMIT transaction ($txn_id) - ERROR: invalid ID; current transaction = ($Meta_txn)\n" if ($METABUG);
	return FALSE;
    }

    &db_reset($dbc);

    my $result = $dbc->exec("COMMIT TRANSACTION");
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
	warn "META COMMIT transaction ($txn_id) - FAIL: " . $dbc->errorMessage . "\n";
	return FALSE;
    }

    warn "META COMMIT transaction ($txn_id) - OK\n" if ($METABUG);
    $Meta_txn = '';

    return TRUE;
}

############### rollback_txn("transaction ID")
#
# syntax:       &rollback_txn($id)
#               abort a transaction
#
sub rollback_txn
{
    my($txn_id) = (shift || $Meta_global_txn);

    if ($Meta_txn eq '')
    {
	warn "META ROLLBACK transaction ($txn_id) - ERROR: no transaction in progress\n" if ($METABUG);
	return FALSE;
    }
    unless ($Meta_txn eq $txn_id)
    {
	warn "META ROLLBACK transaction ($txn_id) - ERROR: invalid ID; current transaction = ($Meta_txn)\n" if ($METABUG);
	return FALSE;
    }

    &db_reset($dbc);

    my $result = $dbc->exec("ROLLBACK TRANSACTION");
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
	warn "META ROLLBACK transaction ($txn_id) - FAIL: " . $dbc->errorMessage . "\n";
	return FALSE;
    }

    warn "META ROLLBACK transaction ($txn_id) - OK\n" if ($METABUG);
    $Meta_txn = '';

    return TRUE;
}

############### setpw()
#
# syntax:       &setpw("dbname", "dbport", "username", "password");
#               defines a password for the user "username" in the database
#
sub setpw
{
    my($dbname,$port,$user,$pw) = (@_);
    my $hba = "/home/pgsql/pg_hba.conf";
    my $hbalock = "/etc/locks/LCK..pg_hba.conf";

    unless ($dbname && $port && $user && $pw) {
	warn "META setpw - ERROR: all arguments must be defined: {dbname,port,user,pw}\n" if ($METABUG);
	return;
    }

    # overwrite the permissions file with unrestricted access for local
    FILE_lock_w($hba, *HBA);
    print HBA "local all trust\n";
    FILE_unlock(*HBA);

    if (defined($dbc)) {
	$dbc->requestCancel;
	$dbc = '';
    }
    my $conn = Pg::connectdb("user=postgres port=$port dbname=$dbname");

    my $command = "ALTER USER $user WITH PASSWORD $pw";
    warn "META setpw - QUERY: $command\n" if ($METABUG);
    my $result = $conn->exec($command);

    # overwrite the permissions file with restricted access for local
    FILE_lock_w($hba, *HBA);
    print HBA "local cobalt crypt\n";
    FILE_unlock(*HBA);

    # write new pw to secure location
    FILE_lock_w($Meta_id, *PGC);
    print PGC "$pw";
    FILE_unlock(*PGC);
    chown 0, 0, $Meta_id;
    chmod 0600, $Meta_id;

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
	warn "META setpw - FAIL: " . $conn->errorMessage . "\n";
	return;
    }

    warn "META setpw - OK: $user\n" if ($METABUG);
    return 1;
}

############### getdb()
#
# syntax:       $dbc = &getdb()
#               returns the database handle
#
sub getdb
{
    return $dbc;
}

############### create_table()
#
# syntax:       $obj->create_table();
#               create a table of type "self->type"
#
sub create_table
{
    my $self = shift;

    unless (defined($self)) {
	warn "META create_table - ERROR: object [self] undefined\n";
	return;
    }

    my $type = $self->gettype;
    unless (defined($type)) {
	warn "META create_table - ERROR: attribute [self->type] undefined\n";
	return;
    }

    my $struct = "";
    while (my($key,$val) = each %$self)
    {
	next if ($key eq "modify");
	if ($key eq "members")
	{
	    # members hash
	    foreach my $member (@{ $$self{members} }) {
		$member->create_table;
	    }
	    next;
	}
	$struct .= " $key ";

	if ($val =~ /^\d+$/) {
	    $struct .= "int4,";
	}
	elsif ($val =~ /^\d+\.\d+$/) {
	    $struct .= "float8,";
	}
	else {
	    $struct .= "text,";
	}
    }
    $struct .= " modify datetime";

    &begin_txn("create_table");

    # attempt to create the sql table
    my $command = "CREATE TABLE $type ($struct);";
    warn "META create_table - QUERY: \"$command\"\n" if ($METABUG);
    my $result = $dbc->exec($command);
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
        warn "META create_table - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("create_table");
	return;
    }

    warn "META create_table - OK\n" if ($METABUG);

    # revoke all permissions from public
    $command = "REVOKE ALL ON $type FROM PUBLIC;";
    warn "META create_table - QUERY: \"$command\"\n" if ($METABUG);
    $result = $dbc->exec($command);
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
        warn "META create_table - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("create_table");
	return;
    }

    # give admin all permissions
    $command = "GRANT ALL ON $type TO $Meta_dbuser;";
    warn "META create_table - QUERY: \"$command\"\n" if ($METABUG);
    $result = $dbc->exec($command);
    unless ($result->resultStatus eq PGRES_COMMAND_OK)
    {
        warn "META create_table - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("create_table");
	return;
    }

    &commit_txn("create_table");
    return $self;
}

############### query()
#
# syntax:       $obj->query("type"     => <type>,
#                           "table"    => <into table>,
#                           "sort"     => <sort field>,
#                           "order"    => "asc" | "dsc",
#                           "distinct" => 1 | 0,
#                           "range"    => [ <low, high> ],
#                           "where"    => [ <left, op, right>,
#                                           <conj, left, op, right>,
#                                           ... ]
#                           )
#
#               query the database with a complex SQL query
#               (more powerful than $obj->retrieve)
#
sub query
{
    my(%kv) = (@_);
    my($query,$in,$w,$result);

    unless (exists($kv{type})) {
	warn "META query - ERROR: attribute [type] unspecified\n";
	return;
    }
    my $type = $kv{type};

    &db_reset($dbc);
    return () unless ($dbc);

    $query  = "SELECT";
    $query .= " DISTINCT" if (exists($kv{distinct}));
    $query .= " *";

    if (exists($kv{keys}))
    {
	if (scalar(@{ $kv{keys} }))
	{
	    chop $query;
	    map { $query .= " $_,"; } @{ $kv{keys} };
	    chop $query;
	}
    }

    $query .= " INTO TABLE $kv{table}" if (exists($kv{table}));
    $query .= " FROM $kv{type}";

    #
    # handle conditions on SQL queries,
    # the $kv{where} key contains arrary of
    #
    # [ <left, op, right>, <conj, left, op, right>, ... ]
    #
    # <...> denotes required grouping
    #
    if (exists($kv{where}))
    {
	my ($conj,$left,$op,$right);
	my $count = 3;
	my @conditions = @{ $kv{where} };
	while (scalar(@conditions) >= $count)
	{
	    if ($count == 3)
	    {
		$count = 4;
		$conj = "WHERE";
	    }
	    else
	    {
		$conj = shift(@conditions);
	    }

	    ($left,$op,$right) = splice(@conditions, 0, 3);
	    $query .= " $conj $left $op '$right'";
	}
    }

    # table sort order determined by kv{sort}
    $query .= " ORDER BY $kv{sort} $kv{order}" if (exists($kv{sort}));

    # select limit determined by kv{limit}
    if (exists($kv{limit}))
    {
	my($count,$offset) = splice(@{ $kv{limit} }, -2);
	$query .= " LIMIT $count OFFSET $offset";
    }

    # select from SQL table
    warn "META query - QUERY: \"$query\"\n" if ($METABUG);
    $result = $dbc->exec($query);
    
    # if we can't find it in the database return
    unless ($result->resultStatus eq PGRES_TUPLES_OK) {
	warn "META query - FAIL: " . $dbc->errorMessage . "\n";
	return;
    }

    if ($result->ntuples < 1) {
	warn "META query - EMPTY: " . $dbc->errorMessage . "\n" if ($METABUG);
	return;
    }

    my (@row,@rows,$tobj,$k);
    if (scalar(@{ $kv{keys} }) == 1)
    {
	# if we only want one key in the result, return a flat array
	while (@row = $result->fetchrow) {
	    push(@rows, $row[0]);
	}
    }
    else
    {
	while (@row = $result->fetchrow)
	{
	    $tobj = Cobalt::Meta->new("type" => "$kv{type}");
	    for ($k=0; $k<$result->nfields; $k++) {
		$tobj->put($result->fname($k) => $row[$k]);
	    }
	    push(@rows, $tobj);
	}
    }

    return(@rows);
}

############### update()
#
# syntax:       $obj->update("type"   => table,
#                            "change" => [ <key, value>, ... ],
#                            "where"  => [ <left, op, right>,
#                                          <conj, left, op, right>,
#                                          ... ]);
#
#               update the <key, value> pairs in the object
#               can pass it limiting arguments in "where" argument
#
sub update
{
    my(%opt)=(@_);

    unless (exists($opt{type}))
    {
	warn "META update - ERROR: attribute [type] must be defined\n";
	return;
    }
    my($type) = $opt{type};

    my(@change) = @{ $opt{change} };
    if (scalar(@change) < 2)
    {
	warn "META update - ERROR: attribute [change] must have at least 2 members\n";
	return;
    }

    my(@where) = @{ $opt{where} };
    if (scalar(@where) < 3)
    {
	warn "META update - ERROR: attribute [where] must have at least 3 members\n";
	return;
    }

    my $update = "UPDATE $type SET";

    #
    # the changes are a list of <key, value> pairs to change
    #
    #              format ex:  [ <key, value>, ... ]
    #
    while (scalar(@change) >= 2)
    {
	my($l,$r) = splice(@change,0,2);
	$update .= " $l = '$r',";
    }
    chop($update);

    $update .= " WHERE";

    #
    # the conditions are a list of <(conj),left,op,right> groups
    # that limit the search for rows to update
    #
    #              format ex:  [ <left, op, right> ]
    #             general ex:  [ <left, op, right>,
    #                            <conj, left, op, right>, ... ]
    #
    my $args = 3;
    while (scalar(@where) >= $args)
    {
	my $c = '';
	if ($args == 3) {
	    $args = 4;
	}
	else {
	    $c = shift(@where);
	}

	my($l,$o,$r) = splice(@where,0,3);
	$update .= " $c $l $o '$r'";
    }

    &begin_txn("update");

    # update SQL table
    warn "META update - QUERY: $update\n" if ($METABUG);
    my $result = $dbc->exec($update);

    if ($result->resultStatus ne PGRES_COMMAND_OK) {
	warn "META update - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("update");
	return;
    }
    
    # return the number of rows updated
    if ($result->cmdStatus =~ /^UPDATE (.+)/) {
	&commit_txn("update");
	return $1;
    }

    warn "META update - EMPTY: " . $dbc->errorMessage . "\n";
    &rollback_txn("update");
    return;
}

############### add_field()
#
# syntax:       add_field("type"  => $table,
#                         "fname" => $key,
#                         "ftype" => $type);
#
sub add_field
{
    my(%a)=(@_);
    unless (exists($a{type}) && exists($a{fname}) && exists($a{ftype})) {
	warn "META add_field - ERROR: undefined 'type' or 'fname' or 'ftype'\n";
	return;
    }

    &begin_txn("add_field");
    my $command = "ALTER TABLE $a{type} ADD COLUMN $a{fname} $a{ftype}";
    warn "META add_field - QUERY: $command\n" if ($METABUG);
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
	warn "META add_field - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("add_field");
	return;
    }

    &commit_txn("add_field");
    return 1;
}

############### retrieve()
#
# syntax:       $obj->retrieve("name", [ <conj, left, op, right>, ... ]);
#               build the object from the SQL database
#
#               can pass it more limiting arguments in
#               [ <conj, left, op, right>, ... ] groups
#
sub retrieve
{
    my ($self,$name,@where) = @_;
    unless ($self && $name) {
	warn "META retrieve - ERROR: [self] and attribute [self->name] must be defined\n";
	return;
    }

    my $query = "SELECT DISTINCT * FROM $$self{type} WHERE name = '$name'";

    my ($c,$l,$o,$r,$where);
    while (scalar(@where) >= 4)
    {
	($c,$l,$o,$r) = splice(@where, 0, 4);
	$query .= " $c $l $o '$r'";
    }

    # select from SQL table
    warn "META retrieve - QUERY: \"$query\"\n" if ($METABUG);
    my $result = $dbc->exec($query);
    
    # if we can't find it in the database return
    unless ($result->resultStatus eq PGRES_TUPLES_OK) {
	warn "META retrieve - FAIL: " . $dbc->errorMessage . "\n" if ($METABUG);
	return;
    }

    # if the query returned nothing return
    unless ($result->ntuples > 0) {
	warn "META retrieve - EMPTY: " . $dbc->errorMessage . "\n" if ($METABUG);
	return;
    }

    # now populate the object
    my(@row) = $result->fetchrow();
    for (my $k = 0; $k < $result->nfields; $k++)
    {
	$self->put($result->fname($k) => $row[$k]);
    }

    return $self;
}

############### get()
#
# syntax:       $value = $obj->get($key);
#               returns the value of $self->{$key} if found,
#               nothing if it can't be found
#
sub get
{
    my ($self,$key) = @_;
    unless (defined($key)) {
	warn "META get - ERROR: no 'key' specified\n";
	return;
    }
    unless (exists($self->{$key})) {
	warn "META get - ERROR: attribute [self->$key] undefined\n" if ($METABUG);
	return;
    }
    return $self->{$key};
}

############### getall()
#
# syntax:       @objects = $obj->getall;
#               returns an array of all object names in the given table
#
sub getall
{
    my ($self,$type) = @_;
    $type = $self->{type} unless defined($type);
    
    # do the query
    my $command = "SELECT * FROM $type";
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_TUPLES_OK) {
	warn "META getall - FAIL: " . $dbc->errorMessage . "\n";
	return;
    }

    unless ($result->ntuples > 0) {
	warn "META getall - FAIL: " . $dbc->errorMessage . "\n";
	return;
    }
    
    my(@all,@row);
    while(@row = $result->fetchrow)
    {
	my $single = Cobalt::Meta->new("type" => "$type");
	for (my $k = 0; $k < $result->nfields; $k++)
	{
	    $single->put($result->fname($k) => $row[$k]);
	}
	push(@all, $single->{name});
    }

    return @all;
}

############### getname()
#
# syntax:       $name = $obj->getname;
#               returns the name of the object
#
sub getname
{
    my $self = shift;
    return $self->{name} if exists($self->{name});

    warn "META getname - ERROR: attribute [self->name] undefined\n";
    return;
}

############### gettype()
#
# syntax:       $type = $obj->gettype;
#               returns the name of the object
#
sub gettype
{
    my $self = shift;
    return $self->{type} if exists($self->{type});

    warn "META gettype - ERROR: attribute [self->type] undefined\n";
    return;
}

############### getmodify()
#
# syntax:       $modify = $obj->getmodify;
#               returns the modification date of the object,
#               formatted per locale
#
sub getmodify
{
    my $self = shift;
    unless (defined($self)) {
	warn "META getmodify - ERROR: object [self] undefined\n";
	return;
    }
    unless (exists($self->{modify})) {
	warn "META getmodify - ERROR: attribute [self->modify] undefined\n";
	return;
    }
    my $modify = $self->{modify};

    # locale-specific date conversion function
    my $convert_funct = $localeObj->getProperty("date_Convert");
    my $convert_module = $localeObj->getProperty("toolsModule");

    # this is how date is stored in the PG database (traditional mode)
    my($d,$m,$h,$n,$s,$y);
    unless ((($d,$m,$h,$n,$s,$y)=($modify =~ /\S+\s+(\d+)\s+(\S+)\s+(\d+):(\d+):(\d+)\s+(\d+)/)) ||
	    (($m,$d,$h,$n,$s,$y)=($modify =~ /\S+\s+(\S+)\s+(\d+)\s+(\d+):(\d+):(\d+)\s+(\d+)/)))
    {
	warn "META getmodify - ERROR: invalid date='$modify'\n";
	return;
    }

    use locale;
    # convert to EPOCH and then back to formatted string
    $modify = strftime("%c %Z", localtime(timelocal($s,$n,$h,$d,$pg_month{$m},$y)));

    # convert the date according to locale properties, if defined
    if ($convert_funct && $convert_module)
    {
	require "$convert_module";
	my $fmodify = &$convert_funct($modify);
	$modify = defined($fmodify) ? $fmodify : $modify;
    }
    no locale;

    return $modify;
}

############### put()
#
# syntax:       $obj->put("key" => "value", ...);
#               insert object (key,value) pairs
#
sub put
{
    my ($self,%pairs) = (@_);

    unless (defined($self)) {
	warn "META put - ERROR: object [self] undefined\n";
	return;
    }

    while (($left,$right) = each %pairs) {
	$self->{$left} = $right;
    }

    return $self;
}

############### delete()
#
# syntax:       $obj->delete();
#               delete object from database
#               based on objects name and type
#
sub delete
{
    my $self = shift;

    unless (defined($self)) {
	warn "META delete - ERROR: object [self] undefined\n";
	return;
    }

    unless (exists($self->{type})) {
	warn "META delete - ERROR: attribute [self->type] undefined\n";
	return;
    }

    unless (exists($self->{name})) {
	warn "META delete - ERROR: attribute [self->name] undefined\n";
	return;
    }

    &begin_txn("delete");
    my $command = "DELETE FROM $$self{type} WHERE name = '$$self{name}'";
    warn "META delete - QUERY: $command\n" if ($METABUG);
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
        warn "META delete - FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("delete");
	return;
    }
    
    &commit_txn("delete");
    return $self;
}

############### vacuum()
#
# syntax:       &vacuum();
#               vacuums database to keep it fast and clean
#
sub vacuum
{
    &db_reset($dbc);

    my $command = "VACUUM ANALYZE";
    warn "META vacuum - QUERY: $command\n" if ($METABUG);
    my $result = $dbc->exec($command);

    unless ($result->resultStatus eq PGRES_COMMAND_OK) {
        warn "META vacuum - FAIL: " . $dbc->errorMessage . "\n";
	return;
    }

    return 1;
}

############### drop()
#
# syntax:       &drop($table, ...);
#               drops tables from the sql database
#
sub drop
{
    my @tables = @_;
    my $ret = 1;

    unless (scalar(@tables) > 0) {
	warn "META drop - ERROR: no tables specified\n";
	return;
    }

    &begin_txn("drop");

    foreach my $table (@tables) {
	my $command = "DROP TABLE $table";
	warn "META drop - QUERY: $command\n" if ($METABUG);
	my $result = $dbc->exec($command);

	unless ($result->resultStatus eq PGRES_COMMAND_OK) {
	    warn "META drop - FAIL: " . $dbc->errorMessage . "\n";
	    $ret = 0;
	}
    }

    if ($ret == 0) {
	&rollback_txn("drop");
	return;
    }

    &commit_txn("drop");

    return 1;
}

############### save()
#
# syntax:       $obj->save();
#               persistent objects by interfacing to a postgresql
#               database, writing the object to table specified by {type}
#
#               all information is dynamically generated from the object
#
sub save
{
    my($self) = shift;
    my($keys,$vals) = ('','');
    my($query,$result);

    ###  check for valid meta data
    unless (defined($self) && exists($self->{type}) && exists($self->{name})) {
	warn "META save - ERROR: object [self] or attributes [self->type],[self->name] undefined\n";
	return;
    }

    ###  build key/value lists for SQL commands
    while (my($key,$val) = each %$self)
    {
	next if ($key eq "modify");

	if ($key eq "members")
	{
	    ###  save recursive members hash
	    foreach my $member (@{ $$self{members} }) {
		$member->save();
	    }
	    next;
	}

	$val =~ s/([\\\'])/\\$1/g;
	$keys .= " $key,";
	$vals .= " '$val',";

	&meta_write($key,$val) if ($METABUG);
    }

    ###  add modification date/time field (as "modify")
    $keys .= " modify";
    $vals .= " datetime('now'::abstime)";

    &begin_txn("save");

    ###  query database to see if any identical tuples exist
    $query = "SELECT DISTINCT * FROM $$self{type} WHERE name = '$$self{name}';";
    warn "META save - QUERY: \"$query\"\n" if ($METABUG);
    $result = $dbc->exec($query);

    if ($result->resultStatus eq PGRES_TUPLES_OK && $result->ntuples > 0)
    {
	###  if we found a matching (conflicting) tuple delete it
	warn "META save - tuple exists: [$$self{type}]->[$$self{name}] - DELETING...\n" if ($METABUG);

	unless ($self->delete) {
	    warn "META save - delete FAIL: [$$self{type}]->[$$self{name}]\n";
	    &rollback_txn("save");
	    return;
	}
    }
    elsif($result->resultStatus eq PGRES_FATAL_ERROR)
    {
	###  this error can result if the table does not exist
	###  so rather than bail completely try to create it
	&rollback_txn("save");
	if ($METABUG) {
	    warn "META save - query FAIL: table does not exist??  trying to create...\n";
	}
	&begin_txn("save");

	unless ($self->create_table())
	{
	    ###  if the table creation fails we have to bail
	    warn "META save - create_table FAIL: " . $dbc->errorMessage . "\n";
	    &rollback_txn("save");
	    return;
	}

	###  and then re-try the query
	warn "META save - QUERY: \"$query\"\n" if ($METABUG);
	$result = $dbc->exec($query);

	###  only care that we get TUPLES_OK from the backend
	unless ($result->resultStatus eq PGRES_TUPLES_OK)
	{
	    ### if we don't get a good query this time then give up
	    warn "META save - create_table FAIL: " .
		 "(" . $result->resultStatus . ") " .
		 $dbc->errorMessage . "exiting...\n";
	    &rollback_txn("save");
	    return;
	}
    }

    ###  now insert the new tuple into the database
    $query = "INSERT INTO $$self{type} ($keys) VALUES ($vals);";
    warn "META save - QUERY: \"$query\"\n" if ($METABUG);
    $result = $dbc->exec($query);

    unless($result->resultStatus eq PGRES_COMMAND_OK) {
	###  if the insert failed then report the error,
	###  rollback the transaction, and bail on this function
	warn "META save - insert FAIL: " . $dbc->errorMessage . "\n";
	&rollback_txn("save");
	return;
    }

    warn "META save - insert OK => OID: " . $result->oidStatus . "\n" if ($METABUG);

    ###  if all is well then commit this
    &commit_txn("save");

    return $self;
}

############### meta_write()
#
# syntax:       meta_write( [<key, value>, ...] );
#               displays formatted (key,value) pairs to STDERR
#
sub meta_write
{
    my(@pairs)=(@_);
    my($META_key,$META_value);

    # save old STDERR handle
    my($old_STDERR) = select(STDERR);

    # define output format
    format STDERR =
@>>>>>>>>>>>>>>>>>>> | ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$META_key,             $META_value
~                      ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<...
                       $META_value
.

    # assign FORMAT_NAME to new STDERR
    $~ = 'STDERR';

    # output each (key,value) pair
    while (($META_key,$META_value) = splice(@pairs, 0, 2))
    {
	write(STDERR);
    }

    # return STDERR handle to original
    select($old_STDERR);

    return;
}

############### display()
#
# syntax:       $obj->display();
#               output the object to STDERR using formatted
#               &meta_write() method
#
sub display
{
    my $self = shift;
    &meta_write(%$self);
}

############### as_xml()
#
# syntax:       $obj->as_xml();
#               returns the object as a simple XML block.
#               any child objects are nested inside it.
#
sub as_xml()
{
    my $self = shift;
    my $depth = shift || 0;
    my $tabs = "\t" x $depth;
    return unless defined($self);
    return unless (my $type = $self->gettype);
    return unless (my $name = $self->getname);
    
    my $xml = "$tabs<$type name=\"$name\">\n";
    foreach my $key (keys %$self)
    {
	next if ($key eq "name");
	next if ($key eq "type");

	if ($key eq "members")
	{
	    foreach (@{ $self->{$key} })
	    {
		$xml .= $_->as_xml( $depth + 1 );
	    }
	}
	else
	{
	    $xml .= "$tabs\t<$key>$$self{$key}</$key>\n";
	}
    }
    $xml .= "$tabs</$type>\n";

    return $xml;
}

###########################################################################
#########################  UNUSED/TESTED METHODS  #########################
###########################################################################

############## from_xml_elem()
#
# syntax:      $obj = &from_xml_elem( $xml_element );
#
sub from_xml_elem()
{
    my ($self,$xml) = (@_);
    return unless defined($xml);

    my $parser = new XML::Parser(Handlers => { Start => \&xml_start,
					       End   => \&xml_end,
					       Char  => \&xml_char });

    # parse calls the xml_* functions, which in turn call the global XML_obj
    $parser->parse($xml);

    return get_xml_obj();
}

############## write_object();
#
# syntax:      $ret = $obj->write_object();
#              if the Object is of a supported type, commit it to disk and return
#              a status string '\d{4} \w+'
#
sub write_object()
{
    my $self = shift;
    return unless defined($self);

    if (exists($write_object_methods{ $$self{type} }))
    {
	return &{ $write_object_methods{ $$self{type} } }($self);
    }

    warn MSG_error("write_object_method__not_supported");
    return;
}

############## populate_members()
#
# syntax:      $obj->populate_members();
#              Objects must have their members specifically requested or added.
#
sub populate_members()
{
    my $self = shift;
    return unless defined($self);

    foreach ($self->get_members)
    {
	$self->add_member($_);
    }
}

############# get_members()
#
# syntax:     @members = $obj->get_members( type=>'vsite', condition=>["ip", "=", "foo"])
# 
sub get_members
{
    my($self,%args) = (@_);
    return unless defined($self);

    unless (exists($args{type}))
    {
	# without a defined type we return all the members we have

	my (@members);
	foreach my $sub_type (keys %{ $member_find{ $$self{type} } })
	{
	    my @cond = @{ $member_find{ $$self{type} }{ $sub_type } };
	    foreach (@cond)
	    {
		# replace the [self.value] lines
		$_ =~ s/\[self\.(\w+)\]/$$self{$1}/ge;
	    }
	    
	    # now grab everything as a query
	    foreach my $obj (Cobalt::Meta::query("type" => $sub_type, "where" => \@cond))
	    {
		# if that object has sub members, we get them too..
		$obj->get_members;
		push (@members, $obj);
	    }
	}
	return @members;
    }
    
    # if we have a type arg, then look for specific members, NOT everything
    my @cond = @{ $member_find{ $$self{type}{ $args{type} } } };

    foreach my $mem (@cond)
    {
	# Replace the [self.value] lines.
	$mem =~ s/\[self\.(\w+)\]/$$self{$1}/ge;
    }

    if (exists($args{where}))
    {
	unshift (@cond, "and");
	push (@{ $args{where} }, @cond);
    }

    # Now we just return whatever the query returns.
    return $self->query(%args);
}

############ get_parents()
#
# syntax:    @parents = $obj->get_parents();
#            who's your daddy?
#
sub get_parents
{
    my $self = shift;
    my ($parent,$qualifiers);
    my @parents;
    
    # This is a bit nasty, but to get it right we have to go through and find
    # out which objects can have objects of our type as children.
    while (($parent, $qualifiers) = each %member_find) {
	if (defined $$qualifiers{$self->gettype}) {
	    my (@condition,$conj,$left,$op,$right);
	    my @cond = @{ $$qualifiers{$self->gettype} };

	    # This should be something like " vsite = '[self.name]' "
	    # we want to turn it into " name = '[self.vsite]'"
	    # Beleive it or not the line below does that.
	    my $arg = 3;
	    while (scalar(@cond) >= $arg) {
		if ($arg == 3) {
		    $conj = '';
		    $arg = 4;
		} else {
		    $conj = shift @cond;
		}
		my $left = shift @cond;
		my $op = shift @cond;
		my $right = shift @cond;
		
		if (defined($conj)) {
		    push @condition, $conj;
		}
		
		$right =~ /\[self\.(\w+)\]/;
		push @condition, $1;
		push @condition, $op;
		push @condition, $self->get($left);
	    }
#			   $qualify =~ s/([^\ ]+)\s+([^\ ]+)\s+\'\[self\.([^\ \]]+)\]\'/
#			   "$3 $2 '".$self->get($1)."'"/gex;
#			   warn "PARENT_SEARCH " if $METABUG; # query()'s debugging will pop
	    # the information about the actual query on for us.
	    # Now we get the objects we want with a new query ;).
	    @parents = Cobalt::Meta::query(type => $parent, "where" => \@condition);
	}
    }
    return @parents;
}


############ dump_data()
#
# syntax:    $obj->dump_data(*FILE);
#            Given a file handle to write to will dump out a tar file
#            of all the data associated with an object
#
sub dump_data
{
    my $self = shift;
    my $fh = shift;
    my ($buffer, $directory, $identifier);
    my $tarcmd = "/bin/tar cz ./";

    # We need some way to know which uids/gids match which objects,
    # site users may own files in a web data area for example.
    print $fh "##### BEGIN COBALT DATA ID MAP #####\n";
    print $fh "uid:$$self{uid}=$$self{name}\n" if (exists($$self{uid}));
    print $fh "gid:$$self{gid}=$$self{name}\n" if (exists($$self{gid}));

    foreach my $member ($self->get_members)
    {
	print $fh "uid:$$member{uid}=$$member{name}\n" if (exists($$member{uid}));
	print $fh "gid:$$member{gid}=$$member{name}\n" if (exists($$member{gid}));
    }

    print $fh "##### END COBALT DATA ID MAP #####\n";
    print $fh "##### BEGIN COBALT DATA FOR $$self{type} $$self{name} #####\n";

    # for the puposes of debugging for now this just exports
    # a lits of the files that *would* be in this data block

    while (($identifier,$directory) = each %{ $files{ $$self{type} } })
    {
	print $fh "##### BEGIN COBALT DATA BLOCK FOR $identifier #####\n";

	$directory =~ s/\[self\.(\w+)\]/$$self{$1}/ge;
	if ($directory =~ /\/$/)
	{
	    # either we're dumping a direcctory
	    if (chdir $directory)
	    {
		open(DATA, "$tarcmd 2> /dev/null |") || return MSG_error("meta_cant_tar");
	    }
	}
	else
	{
	    # or we're dumping a file (missing ok)
	    open(DATA, "<$directory");
	}

	while (read(DATA, $buffer, 57)) {
	    print $fh encode_base64($buffer);
	}

	print "##### END COBALT DATA BLOCK FOR $identifier #####\n";
    }
}

############ restore_data()
#
# syntax:    $obj->(*FILE);
#            restore data from filehandle
#
sub restore_data
{
    my $self = shift;
    my $fh = shift;

    my $line;
    my $location;

    my $tarcmd = '/bin/tar x';
    
    while (<$fh>) {
	if (/##### BEGIN COBALT DATA BLOCK FOR (.*) #####/o) {
	    $location = $1;
	    if ( $files{$self->gettype}{$location} =~ /\/$/ ) {
		#open(DATA, "|$tarcmd -C $location") ||
		#	return MSG_error("meta_cant_tar");
		warn "Untarring to -C " . $files{$self->gettype}{$location} . "\n";
		open(DATA, "|/bin/tar tz 1>&2");
	    } elsif ( $files{$self->gettype}{$location} =~ /[^\/]$/) {
		#open(DATA, ">$location") ||
		#	warn MSG_error("meta_cant_write");
		warn "Dumping file to $location.\n";
	    } else {
		# Eeek! No idea where to dump this.
		warn MSG_error("meta_unknown_data_identifier");
	    }
	    # bloody perl.
	    $line = <$fh>;
	    while( $line !~ /##### END COBALT DATA BLOCK .* #####/o ) {
		   print DATA decode_base64($line);
		   $line = <$fh>;
	       }
	}
    }
}

############ dump_object()
#
# syntax:    $obj->dump_object(*FILE);
#
sub dump_object
{
    my $self = shift;
    my $fh = shift;

    print "##### BEGIN COBALT OBJECT #####\n";

	print $fh $self->as_xml();

    $self->dump_data($fh);

    foreach ($self->{members}) {
	print $fh $_->dump_object();
    }
}

############ add_member()
#
# syntax:    $obj->add_member($newobj);
#
sub add_member
{
    my $self = shift;
    my $new_obj = shift;
    unless (defined $self->{members}) {
	$self->{members} = [];
    }
    push @{ $$self{members} }, $new_obj;
}

# Our xml parsing handlers and variables they work with live here in their
# own block so as to minimise name space polution.
my @swap;
my $xml_obj;
my $type;

############ xml_start()
#
#            first time we see a 'start' element,
#            it's for what type of object this is
#
sub xml_start {
    my $parser = shift;
    my $section = shift;
    my %args = @_;
    # Only tags of type "<type name="name">" start of blocks.
    if ( $args{name} ) {
	if ( defined $$xml_obj{type} ) {
	    # We currently have an object being pricessed, dump it on the stack
	    # to be picjed up later.
	    #warn "Pushing object onto stack\n";
	    push @swap, $xml_obj;
	}
	#warn "Creating $section object with";
	#while (my($key) = each(%args)) {
	#	print ", $key = $args{$key}";
	#}
	#print "\n";
	$xml_obj = Cobalt::Meta->new(type => $section,
				     name => $args{name});
	return;
    }
}

############ xml_end()
#
sub xml_end {
    my $parser = shift;
    my $section = shift;
    if ( $section eq $$xml_obj{type} ) {

	# If we're ending a section for the type of object we're
	# currentl parsing we're done with this block.
	if ( scalar( @swap ) ) {

	    # O.K. We were parsing a sub block.
	    # Pop our parent block off the stack..
	    my $swap = pop @swap;

	    # Add the current block as a member of the parent block.
	    $swap->add_member($xml_obj);

	    # And make the parent block the current so we can continue where we left off.
	    $xml_obj = $swap;
	}
	# If we were at the top of the block we do nothing.
    }
}

############ xml_char()
#
sub xml_char {
    my $parser = shift;
    my $section = shift;
    # Skip white space.
    if ( $section =~ /^\s+$/) {
	return;
    }
    $xml_obj->put($parser->current_element(), $section);
    #print "Setting ".$parser->current_element()." to $section\n";
}

############ get_xml_obj()
#
# syntax:    $XMLobject = $obj->get_xml_obj()
#
sub get_xml_obj {
    return $xml_obj;
}

###########################################################################
##########################  DEPRECATED METHODS  ###########################
###########################################################################

############### create()
#
# syntax:       $obj = Cobalt::Meta->create("type" => $type);
#               returns an empty object belonging to database $type
#
# syntax:       $obj = Cobalt::Meta->create("type" => $type,
#                                           "key"  => $value, ...);
#               returns an object with corresponding ("key" => $value) traits
#
#               DEPRECATED:  use new()
#
sub create
{
    my $data = shift;
    my $obj=new($data, @_);
    return $obj;
}

############### add_column()
#
# syntax:       &add_column("type"    => $type,
#                           "colname" => $key,
#                           "coltype" => $valtype);
#
#               DEPRECATED:  see &add_field()
#
sub add_column
{
    my %arg = @_;

    return unless (exists ($arg{type}) &&
		   exists ($arg{colname}) &&
		   exists ($arg{coltype}));

    return Cobalt::Meta::add_field("type"  => "$arg{type}",
				   "fname" => "$arg{colname}",
				   "ftype" => "$arg{coltype}");
}

1;

__END__

